"""Encoding and decoding blocks of tokun."""

import keras
import tensorflow as tf

import mlable.layers.reshaping
import mlable.layers.transformer

# ENCODING BLOCKS #############################################################

@keras.saving.register_keras_serializable(package='blocks')
class TokenizeBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        sequence_axis: int=0,
        feature_axis: int=-1,
        token_dim: int=4,
        embedding_dim: int=256,
        activation: str='gelu',
        epsilon: float=1e-6,
        **kwargs
    ) -> None:
        super(TokenizeBlock, self).__init__(**kwargs)
        # config
        self._config = {
            'sequence_axis': sequence_axis,
            'feature_axis': feature_axis,
            'token_dim': token_dim,
            'embedding_dim': embedding_dim,
            'activation': activation,
            'epsilon': epsilon,}
        # layers
        self._normalize = tf.keras.layers.LayerNormalization(axis=feature_axis, epsilon=epsilon, center=True, scale=True, name='normalization') # normalize each token unit independently
        self._divide = mlable.layers.reshaping.Divide(input_axis=sequence_axis, output_axis=feature_axis, factor=token_dim, insert=False, name='reshaping') # (B, S * G, E) => (B, S, G * E)
        self._dense = tf.keras.layers.Dense(units=embedding_dim, activation=activation, use_bias=True, kernel_initializer='glorot_uniform', bias_initializer='zeros', name='compression') # (B, S, G * E) => (B, S, L), typically L = E

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return self._dense(self._divide(self._normalize(inputs)))

    def get_config(self) -> dict:
        __config = super(TokenizeBlock, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# DECODING BLOCKS #############################################################

@keras.saving.register_keras_serializable(package='blocks')
class DetokenizeBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        sequence_axis: int=0,
        feature_axis: int=-1,
        token_dim: int=4,
        embedding_dim: int=256,
        activation: str='gelu',
        epsilon: float=1e-6,
        **kwargs
    ) -> None:
        super(DetokenizeBlock, self).__init__(**kwargs)
        # config
        self._config = {
            'sequence_axis': sequence_axis,
            'feature_axis': feature_axis,
            'token_dim': token_dim,
            'embedding_dim': embedding_dim,
            'activation': activation,
            'epsilon': epsilon,}
        # layers
        self._dense = tf.keras.layers.Dense(units=token_dim * embedding_dim, activation=activation, use_bias=True, kernel_initializer='glorot_uniform', bias_initializer='zeros', name='decompression') # (B, S, L) => (B, S, G * E), typically L = E
        self._divide = mlable.layers.reshaping.Divide(input_axis=feature_axis, output_axis=sequence_axis, insert=False, factor=token_dim, name='reshaping') # (B, S, G * E) => (B, S * G, E)
        self._normalize = tf.keras.layers.LayerNormalization(axis=feature_axis, epsilon=epsilon, center=True, scale=True, name='normalization') # normalize each token unit independently

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return self._normalize(self._divide(self._dense(inputs)))

    def get_config(self) -> dict:
        __config = super(DetokenizeBlock, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# HEAD ########################################################################

@keras.saving.register_keras_serializable(package='blocks')
class HeadBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        encoding_dim: int=256,
        activation: str='softmax',
        **kwargs
    ) -> None:
        super(HeadBlock, self).__init__(**kwargs)
        # config
        self._config = {'encoding_dim': encoding_dim, 'activation': activation}
        # layers
        self._dense = tf.keras.layers.Dense(units=encoding_dim, activation=activation, use_bias=True, kernel_initializer='glorot_uniform', bias_initializer='zeros', name='projection') # (..., G, E) => (..., G, U), typically U = E

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return self._dense(inputs)

    def get_config(self) -> dict:
        __config = super(HeadBlock, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)



"""Piece together the actual VAE CNN model for tokun."""

import functools

import keras
import tensorflow as tf

import tokun.layers

# ENCODER #####################################################################

@keras.saving.register_keras_serializable(package='models')
class Encoder(tf.keras.models.Model):
    def __init__(
        self,
        token_dim: list,
        encoding_dim: int,
        embedding_dim: int,
        sequence_axis: int=1,
        feature_axis: int=-1,
        activation: str='gelu',
        **kwargs
    ) -> None:
        # init
        super(Encoder, self).__init__(**kwargs)
        # config
        self._config = {
            'token_dim': token_dim,
            'encoding_dim': encoding_dim,
            'embedding_dim': embedding_dim,
            'sequence_axis': sequence_axis,
            'feature_axis': feature_axis,
            'activation': activation,}
        # successive dimensions of the merging units
        __token_dim = [token_dim] if isinstance(token_dim, int) else token_dim
        # layers
        self._layers = [
            # (B * G ^ D, U) => (B * G ^ D, E)
            tf.keras.layers.Embedding(
                input_dim=encoding_dim,
                output_dim=embedding_dim,
                embeddings_initializer='glorot_uniform',
                name='embed-1'),] + [
            # (B * G ^ i, E) => (B * G ^ (i-1), E)
            tokun.layers.TokenizeBlock(
                sequence_axis=sequence_axis,
                feature_axis=feature_axis,
                token_dim=__g,
                embedding_dim=embedding_dim,
                activation=activation,
                name='tokenize-{}_{}'.format(__g, __i))
            for __i, __g in enumerate(__token_dim)]

    def call(self, x: tf.Tensor) -> tf.Tensor:
        return functools.reduce(lambda __x, __l: __l(__x), self._layers, x)

    def get_config(self) -> dict:
        __config = super(Encoder, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# DECODER #####################################################################

@keras.saving.register_keras_serializable(package='models')
class Decoder(tf.keras.models.Model):
    def __init__(
        self,
        token_dim: list,
        encoding_dim: int,
        embedding_dim: int,
        sequence_axis: int=1,
        feature_axis: int=-1,
        activation: str='gelu',
        output: str='categorical',
        **kwargs
    ) -> None:
        # init
        super(Decoder, self).__init__(**kwargs)
        # config
        self._config = {
            'token_dim': token_dim,
            'encoding_dim': encoding_dim,
            'embedding_dim': embedding_dim,
            'sequence_axis': sequence_axis,
            'feature_axis': feature_axis,
            'activation': activation,
            'output': output,}
        # successive dimensions of the dividing units
        __token_dim = [token_dim] if isinstance(token_dim, int) else token_dim
        # binary vs categorical probabilities
        __activation = 'softmax' if output == 'categorical' else 'sigmoid'
        # layers
        self._layers = [
            # (B * G ^ i, E) => (B * G ^ (i+1), E)
            tokun.layers.DetokenizeBlock(
                sequence_axis=sequence_axis,
                feature_axis=feature_axis,
                token_dim=__g,
                embedding_dim=embedding_dim,
                activation=activation,
                name='detokenize-{}_{}'.format(__g, __i))
            for __i, __g in enumerate(__token_dim)] + [
            # (B * G ^ D, E) => (B * G ^ D, U)
            tokun.layers.HeadBlock(encoding_dim=encoding_dim, activation=__activation, name='project-head')]

    def call(self, x: tf.Tensor) -> tf.Tensor:
        return functools.reduce(lambda __x, __l: __l(__x), self._layers, x)

    def get_config(self) -> dict:
        __config = super(Decoder, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)

# VAE #########################################################################

@keras.saving.register_keras_serializable(package='models')
class AutoEncoder(tf.keras.models.Model):
    def __init__(
        self,
        token_dim: list,
        input_dim: int,
        output_dim: int,
        embedding_dim: int,
        sequence_axis: int=1,
        feature_axis: int=-1,
        activation: str='gelu',
        output: str='categorical',
        **kwargs
    ) -> None:
        # init
        super(AutoEncoder, self).__init__(**kwargs)
        # config
        self._config = {
            'token_dim': token_dim,
            'input_dim': input_dim,
            'output_dim': output_dim,
            'embedding_dim': embedding_dim,
            'sequence_axis': sequence_axis,
            'feature_axis': feature_axis,
            'activation': activation,
            'output': output,}
        # layers
        self._encoder = Encoder(token_dim=token_dim, encoding_dim=input_dim, embedding_dim=embedding_dim, sequence_axis=sequence_axis, feature_axis=feature_axis, activation=activation)
        self._decoder = Decoder(token_dim=token_dim[::-1], encoding_dim=output_dim, embedding_dim=embedding_dim, sequence_axis=sequence_axis, feature_axis=feature_axis, activation=activation, output=output)

    def call(self, x: tf.Tensor) -> tf.Tensor:
        return self._decoder(self._encoder(x))

    def get_config(self) -> dict:
        __config = super(AutoEncoder, self).get_config()
        __config.update(self._config)
        return __config

    @classmethod
    def from_config(cls, config) -> tf.keras.layers.Layer:
        return cls(**config)



import tensorflow as tf

# CCE #########################################################################

class CategoricalCrossentropyFromEmbeddings(tf.keras.losses.CategoricalCrossentropy):
    def __init__(self, decoder: callable, name='categorical_crossentropy_from_embeddings', **kwargs) -> None:
        # init
        super(CategoricalCrossentropyFromEmbeddings, self).__init__(from_logits=False, label_smoothing=0.0, axis=-1, reduction='sum_over_batch_size', **kwargs)
        # decoder
        self._decoder = decoder

    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        # decode
        __yt = self._decoder(y_true)
        __yp = self._decoder(y_pred)
        # decompression factor
        __dim = int(tf.size(__yt) / tf.size(y_true))
        # expanded loss
        __loss = super(CategoricalCrossentropyFromEmbeddings, self).call(y_true=__yt, y_pred=__yp)
        # expanded shape
        __shape = list(__loss.shape)
        # group by token
        __shape[-1] = __shape[-1] // __dim
        __shape.append(__dim)
        # reshape
        __loss = tf.reshape(__loss, shape=__shape)
        # reduce so that the loss shape matches the input shapes
        return tf.reduce_mean(__loss, axis=-1)

    def get_config(self) -> dict:
        __config = super(CategoricalCrossentropyFromEmbeddings, self).get_config()
        __config.update({'decoder': self._decoder})
        return __config

    @classmethod
    def from_config(cls, config: dict) -> tf.keras.losses.Loss:
        __decoder = config.pop('decoder')
        return cls(decoder=__decoder, **config)



"""Pre and post processing pipelines for tokun."""

import functools
import itertools
import math

import tensorflow as tf

import mlable.ops
import mlable.sampling
import mlable.utils

# UNICODE #####################################################################

CODE_STX = b'\x02'
CODE_ETX = b'\x03'
CODE_FS = b'\x1c'
CODE_GS = b'\x1d'
CODE_RS = b'\x1e'
CODE_US = b'\x1f'

# ENCODE ######################################################################

def encode(data: tf.Tensor, token_size: int, sample_size: int, dtype: tf.dtypes.DType=tf.dtypes.int32) -> tf.Tensor:
    # factor 4 because of the UTF-32 encoding
    __dim = math.ceil(4 * sample_size / token_size) * token_size
    # decode bytes from UTF-8
    __bytes = tf.strings.unicode_transcode(input=data, input_encoding='UTF-8', output_encoding='UTF-32-BE') # (B,)
    # decode byte strings to arrays of byte integers
    __bytes = tf.io.decode_raw(__bytes, out_type=tf.uint8, fixed_length=__dim) # (B, 4 * S)
    # cast to int32 as uint8 is not std
    return tf.cast(__bytes, dtype=dtype)

# RESHAPE #####################################################################

def chunk(seq: list, size: int, repeats: bool=True) -> list:
    __chunks = (seq[__i:__i+size] for __i in range(0, len(seq), size))
    return list(__chunks if repeats else set(__chunks))

def merge(chunks: list) -> list:
    return list(itertools.chain.from_iterable(chunks))

def shape(expand: list=[]) -> list:
    return expand + [-1]

def reshape(data: tf.Tensor, expand: list=[]) -> tf.Tensor:
    # group by token unit
    __shape = shape(expand=expand)
    # partition or flatten the data
    return tf.reshape(tensor=data, shape=__shape) # for example (-1, G, G, G) the first dimension is not B

# AUGMENT #####################################################################

def offset(data: tf.Tensor, ticks: int=1) -> tf.Tensor:
    return tf.convert_to_tensor([ticks * b'\x00']) + data

# DECODE ######################################################################

def decode(data: tf.Tensor) -> str:
    # make sure the dtype is large enough for UTF-32 codepoints
    __data = tf.cast(data, dtype=tf.dtypes.int32)
    # group the bytes 4 by 4
    __shape = mlable.utils.divide_shape(shape=__data.shape, input_axis=-2, output_axis=-1, factor=4, insert=True)
    __bytes = tf.reshape(tensor=__data, shape=__shape)
    # compute the UTF-32-BE codepoints
    __codes = mlable.ops.reduce_base(data=__bytes, base=256, axis=-1, keepdims=False)
    # actually decode
    __utf32 = tf.strings.unicode_encode(__codes, output_encoding='UTF-32-BE')
    # convert to standard UTF-8
    return tf.strings.unicode_transcode(input=__utf32, input_encoding='UTF-32-BE', output_encoding='UTF-8')

# > ###########################################################################

def preprocess(text: str, token_size: int, expand: list=[]) -> tf.Tensor:
    # as tensor
    __data = tf.convert_to_tensor(text, dtype=tf.dtypes.string)
    # list of bytes
    __bytes = encode(data=__data, token_size=token_size, sample_size=len(text))
    # partition or flatten
    return reshape(data=__bytes, expand=expand)

# < ###########################################################################

def unpad(text: str) -> str:
    return text.strip('\x00')

def unpack(data: tf.Tensor) -> list:
    __data = data.numpy().tolist()
    return [__s.decode('utf-8') for __s in __data]

def postprocess(prediction: tf.Tensor, binary: bool=False, random: bool=False) -> str:
    # from one-hot to UTF-32 bytes
    __output = mlable.sampling.binary(prediction=prediction, threshold=0.5, random=random) if binary else mlable.sampling.categorical(prediction=prediction, random=random)
    # flatten the groups of 4 bytes
    return decode(data=__output)

# SAMPLING ####################################################################

def sample(model: tf.keras.models.Model, text: str, **kwargs) -> tuple:
    __x = preprocess(text=text, token_size=kwargs.get('token_size', 16), expand=kwargs.get('expand', [1]))
    __e = model._encoder(__x)
    __p = model._decoder(__e)
    __y = postprocess(__p, binary=kwargs.get('binary', False), random=kwargs.get('random', False))
    __o = unpack(__y)
    return (__x, __e, __p, __y, __o)



import copy
import json

import transformers

# UTF-32-BE TOKENIZATION ######################################################

class ByteTokenizer(transformers.PreTrainedTokenizer):
    def __init__(self, encoding: str='utf-32-be', **kwargs) -> None:
        __kwargs = copy.deepcopy(kwargs)
        # special tokens
        __kwargs['bos_token'] = __kwargs.get('bos_token', '\u0002')
        __kwargs['eos_token'] = __kwargs.get('eos_token', '\u0003')
        __kwargs['unk_token'] = __kwargs.get('unk_token', '\u001a')
        __kwargs['sep_token'] = __kwargs.get('sep_token', '\u001d')
        __kwargs['pad_token'] = __kwargs.get('pad_token', '\u0000')
        __kwargs['cls_token'] = __kwargs.get('cls_token', '\u0011')
        # defaults
        __kwargs['split_special_tokens'] = True # 4 bytes for the special tokens too
        __kwargs['vocab_size'] = 256
        # properties
        self._vocab_size = 256
        self._encoding = encoding
        # init
        super(ByteTokenizer, self).__init__(**__kwargs)

    def _tokenize(self, text: str, **kwargs) -> list:
        return list(chr(__b) for __b in text.encode(self._encoding))

    def _convert_token_to_id(self, token: str) -> int:
        return ord(token)

    def _convert_id_to_token(self, index: int) -> str:
        return chr(index)

    def convert_tokens_to_string(self, tokens: iter) -> str:
        return bytes(ord(__c) for __c in tokens).decode(self._encoding)

    def build_inputs_with_special_tokens(self, token_ids_0: list, token_ids_1: list=None) -> list:
        __cls_ids = [self._convert_token_to_id(__t) for __t in self.tokenize(self.cls_token, split_special_tokens=True)]
        __sep_ids = [self._convert_token_to_id(__t) for __t in self.tokenize(self.sep_token, split_special_tokens=True)]
        __ids = __cls_ids + token_ids_0 + __sep_ids
        return __ids + token_ids_1 + __sep_ids if token_ids_1 else __ids

    def get_vocab(self) -> dict:
        return {chr(__i): __i for __i in range(self._vocab_size)}

    def save_vocabulary(self, save_directory: str, **kwargs) -> tuple:
        __prefix = kwargs.get('filename_prefix', '')
        __path = "{}/{}vocab.json".format(save_directory, __prefix if __prefix else '')
        with open(__path, "w") as __file:
            json.dump(self.get_vocab(), __file)
        return (__path,)

    @property
    def vocab_size(self) -> int:
        return self._vocab_size

    @vocab_size.setter
    def vocab_size(self, size: int) -> None:
        self._vocab_size = size



"""Evaluate the quality of the model output."""

import functools
import itertools
import math

import tensorflow as tf

# ACCURACY ####################################################################

def compare(left: str, right: str) -> float:
    return sum(__l == __r for __l, __r in zip(left, right)) / max(1, len(left))

# TOKEN CONTENT ###############################################################

def intersection(left: str, right: str) -> float:
    __intersection = len(set(left).intersection(set(right)))
    __reference = min(len(set(left)), len(set(right)))
    return __intersection / max(1, __reference)

# ROBUSTNESS ##################################################################

def neighbors(point: tf.Tensor, radius: float, count: int) -> tf.Tensor:
    return point + tf.random.uniform(shape=(count, point.shape[-1]), minval=-radius, maxval=radius, dtype=point.dtype)



"""Setup the hyper parameters for tokun."""

def rates(pretrained: bool=False, normalization: bool=True, base: float=0.001) -> tuple:
    return (
        (0.1 if pretrained else 1.) * (0.1 if normalization else 1.) * base, # lr max
        0.1, # lambda => global decay
        0.9, # beta_1 => decay for the first moment
        0.99) # beta_2 => decay for the second moment

def version(token_units: list, sequence_axis: int=1, input_dim: int=256, output_dim: int=256) -> list:
    return ['{}x{}'.format(input_dim, output_dim), 'x'.join(str(__u) for __u in token_units), str(sequence_axis)]
